# Peoduction class
#
# The foundation of docgen. Everything is based on this class.

# Imports {{{1
from __future__ import print_function, division
from .preferences import (
    DEFAULT_INDENT, DEFAULT_LEADERS, STRIP_TRAILING_WHITESPACE
)
import collections
import re
from six import string_types


# Globals {{{1
# do not import these codes, if you need them, use with control code functions
CTRL_CODE = '\a'
NEWLINE_CODE = CTRL_CODE + 'n'
TAB_CODE = CTRL_CODE + 't'
KILL_CODE = CTRL_CODE + 'k'
INDENT_CODE = CTRL_CODE + 'i'
LEADER_CODE = CTRL_CODE + 'l'
DEDENT_CODE = CTRL_CODE + 'd'
DEFERRED_INDENT_CODE = CTRL_CODE + 'I'
DEFERRED_LEADER_CODE = CTRL_CODE + 'L'
DEFERRED_DEDENT_CODE = CTRL_CODE + 'D'
EMPTY_CODE = CTRL_CODE + 'e'


# Control Codes {{{1
def newlineCode(blankLines=0):
    """Insert newlines

    Adjacent newline codes are combined with the resulting number of blank lines 
    being the maximum specified on any of the adjacent codes. By default, the 
    number of blank lines is 0.  Use blankLines='!' to defeat the combining of 
    newlines.

    :param blankLines: The desired number of blank lines.
    """
    return NEWLINE_CODE + str(blankLines)

def tabCode(column):
    """Inserts space until the specified column is reached

    :param column: The column of the desired tabstop.
    """
    return TAB_CODE + str(column)

def killCode():
    """Strikes the current line"""
    return KILL_CODE

def indentCode(levels=1, deferred=False):
    """Start adding indent to each line

    :param levels:
        The desired number of indent levels.
    :param deferred:
        The indent takes effect on the next line if deferred is True and the 
        current line otherwise.
    """
    if deferred:
        return DEFERRED_INDENT_CODE + str(levels)
    else:
        return INDENT_CODE + str(levels)

def leaderCode(key, deferred=False):
    """Start adding leader to each line

    :param key:
        The key for the desired leader must be specified.
    :param deferred:
        The indent takes effect on the next line if deferred is True and the 
        current line otherwise.
    """
    if deferred:
        return DEFERRED_LEADER_CODE + str(key)
    else:
        return LEADER_CODE + str(key)

def dedentCode(levels=1, deferred=False):
    """Discard one or more levels of indents or leaders

    :param levels:
        The desired number of indents or leaders to discard.
    :param deferred:
        The indent takes effect on the next line if deferred is True and the 
        current line otherwise.
    """
    if deferred:
        return DEFERRED_DEDENT_CODE + str(levels)
    else:
        return DEDENT_CODE + str(levels)

def emptyCode():
    """No-op.

    Use this when you need to specify a rule but want DocGen to act like you did 
    not. The distinction between this and an empty string is significant when 
    processing Newlines(). Two Newlines() separated by an emptyCode() will be 
    combined, where as they will not if they are separated by an empty string.
    """
    return EMPTY_CODE

# Utilities {{{1
# strip_trailing_whitespace() {{{2
if STRIP_TRAILING_WHITESPACE:
    def strip_trailing_whitespace(text):
        return '\n'.join([line.rstrip(' \t') for line in text.splitlines()])

        # the above strips any trailing newline, the following does not
        #return (
        #    '\n'.join([line.rstrip(' \t') for line in text.splitlines()]) +
        #    '\n' if text[-1] == '\n' else ''
        #)
else:
    def strip_trailing_whitespace(text):
        return text

# is_str() {{{2
def is_str(obj):
    return isinstance(obj, string_types)

# is_iterable {{{2
def is_iterable(obj):
    return isinstance(obj, collections.Iterable)


# Production class {{{1
class Production(object):
    # Description {{{2
    """A Description of a Syntax Fragment

    A Production takes a rule as input, and it converts that rule to a string 
    when the render() method is called or when a production object is cast to 
    a string using the str() operator.

    :param rule:
        A rule may be either a string, a Production object, or a list of 
        rules.  In this way a rule may be a hierarchical object of arbitrary 
        depth.  If a rule is None, or contains a rule that is None, then the 
        production is not active, meaning that it will not generate any 
        output.
    """

    # Constructor {{{2
    def __init__(self, rule):
        #assert Production._eligibleToBeRule(rule), ('Bad type: %s' % type(rule))
        self.add_rule(rule)

    # add_rule() {{{2
    def add_rule(self, rule):
        """Add rule

        Used by subclasses to add the rule to the production.
        """
        self.rule = self._clean(rule)

    # _clean() {{{2
    @staticmethod
    def _clean(self, cullArrays = True):
        # Returns its argument if the argument does not evaluate to None. It
        # evaluates to None if it is None, or if it is iterable and one of its
        # members is None.
        # Is a static method because self may not be instance of Production.
        if hasattr(self, 'docgen_rule'):
            assert not hasattr(self, 'docgen_template')
            assert not hasattr(self, 'docgen_templet')
            return Production._clean(self.docgen_rule())
        if hasattr(self, 'docgen_templet'):
            from .templet import Templet
            if isinstance(self.docgen_templet, collections.Callable):
                self = Templet._elaborate(self, self.docgen_templet())
            else:
                self = Templet._elaborate(self, self.docgen_templet)
        if hasattr(self, 'docgen_template'):
            from .template import Template
            symbols = getattr(self, 'docgen_symbols', None)
            if isinstance(self.docgen_template, collections.Callable):
                self = Template._elaborate(self, self.docgen_template(), symbols=symbols)
            else:
                self = Template._elaborate(self, self.docgen_template, symbols=symbols)
        if is_str(self):
            return self
        if isinstance(self, Production):
            return self if self.rule is not None else None
        if is_iterable(self):
            self = [Production._clean(each) for each in self]
            if cullArrays:
                return self if None not in self else None
            else:
                return self
        assert self is None, '%s has unexpected type: %s' % (self, type(self))
        return None

    # _flatten() {{{2
    @staticmethod
    def _flatten(self):
        # Flatten rule to produce a stream of tokens, where tokens are either 
        # control codes, strings, or Nones.
        # Is a static method because self may not be instance of Production.
        if is_str(self):
            # Remove any newlines off the ends of strings, they should be
            # specified explicitly with Newline().
            # Then split the string at any remaining newlines and yield each
            # segment individually while yielding a newline code between each
            # segment.
            segments  = self.strip('\n').split('\n')
            for each in segments[:-1]:
                assert CTRL_CODE not in each
                yield each
                yield NEWLINE_CODE + '!'
            yield segments[-1]
        elif is_iterable(self):
            for each in self:
                for sub in Production._flatten(each):
                    yield sub
        elif isinstance(self, Production):
            for each in Production._flatten(self.rule):
                yield each
        elif hasattr(self, 'docgen_rule') or hasattr(self, 'docgen_template'):
            for each in Production._flatten(Production._clean(self)):
                yield each
        elif self is not None:
            raise AssertionError("Bad type %s" % type(self))

    # _strikeLines() {{{2
    def _strikeLines(self):
        # partition a stream of tokens into groups that represent lines, and 
        # then kill those groups that contain line-kill codes. The first output 
        # line is treated specially, because output is assumed to start on 
        # a newline. Normally, a request for a blank line would result in two 
        # newlines being output, one to terminate the prior line, and one to 
        # create the blank line. But the first line is special because we do not 
        # need to terminate the prior line. Normally, we expect the user to know 
        # this and act accordingly, but things break down when the first lines 
        # are killed. To assure that we do not output extra newlines if lines 
        # are killed before there is any output, the newlineNeeded flag is 
        # initially set to False, and then set true once something has been 
        # output.
        line = []
        kill = False
        firstLine = True
        for token in self._flatten(self.rule):
            if token and token[0] == CTRL_CODE:
                code = token[0:2]
                if code == NEWLINE_CODE:
                    arg = token[2:]
                    if kill:
                        line = []
                        try:
                            newlines = int(arg)
                            if newlines >= 1:
                                if firstLine:
                                    newlines -= 1
                                yield NEWLINE_CODE + str(newlines)
                        except ValueError:
                            # this is a traditional newline (code=!), kill it too
                            pass
                    else:
                        for each in line:
                            if each and each[0] != CTRL_CODE:
                                firstLine = False
                            yield each
                        yield token
                        line = []
                    kill = False
                elif code == KILL_CODE:
                    kill = True
                elif code in [
                    INDENT_CODE, LEADER_CODE, DEDENT_CODE, DEFERRED_INDENT_CODE,
                    DEFERRED_LEADER_CODE, DEFERRED_DEDENT_CODE,
                ]:
                    # Pass these code through regardless of whether then line is 
                    # being killed, and do it now, their order is not important.
                    yield token
                else:
                    line.append(token)
            else:
                line.append(token)
        for each in line:
            yield each

    # _filter() {{{2
    def _filter(self, unitIndent, leaders):
        """
        Convert production to a string while filtering out suppressed sequences
        and extra newlines, and adding indenting.
        """
        column = 0
        indents = []
        deferredIndents = []
        deferredDedents = 0
        newlines = 0
        prev_newline_was_soft = False
        line = ''

        for item in self._strikeLines():
            if item and item[0] == CTRL_CODE:
                code = item[0:2]
                if code == NEWLINE_CODE:
                    arg = item[2:]
                    if not newlines:
                        yield ''.join(indents) + line
                        line = ''
                        column = sum([len(indent) for indent in indents])

                    if arg == '!': # this is a hard newline
                        if prev_newline_was_soft:
                            # if previous newline was soft, convert it to hard 
                            # but do not increment it.
                            prev_newline_was_soft = False
                        else:
                            newlines = newlines + 1
                    else:          # this is a soft newline
                        newlines = max(newlines, int(arg) + 1)
                        prev_newline_was_soft = True

                    # Process any deferred indents and dedents
                    if deferredIndents:
                        indents += deferredIndents
                        deferredIndents = []
                    if deferredDedents:
                        assert len(indents) >= deferredDedents, (
                            'too many deferred dedents'
                        )
                        indents = indents[:-deferredDedents]
                        deferredDedents = 0

                elif code == TAB_CODE:
                    tab = int(item[2:])
                    if column > tab:
                        # User has requested a tab, but the current column is 
                        # beyond the tabstop. There are two different ways this 
                        # can be handled:
                        # 1. either emit a linefeed and then tab to tabstop
                        padding = '\n' + tab * ' '
                        column = tab
                        # 2. or simple add a space
                        #padding = ' '
                        #column += 1
                    else:
                        padding = (tab - column) * ' '
                        column = tab
                    line += padding
                elif code == KILL_CODE:
                    raise AssertionError
                elif code == INDENT_CODE:
                    levels = int(item[2:])
                    indents += levels*[unitIndent]
                    column = sum([len(indent) for indent in indents])
                elif code == LEADER_CODE:
                    key = item[2:]
                    indents += [leaders[key]]
                elif code == DEDENT_CODE:
                    levels = int(item[2:])
                    assert len(indents) >= levels, 'too many dedents'
                    indents = indents[:-levels]
                elif code == DEFERRED_INDENT_CODE:
                    levels = int(item[2:])
                    deferredIndents += levels*[unitIndent]
                elif code == DEFERRED_LEADER_CODE:
                    key = item[2:]
                    deferredIndents += [leaders[key]]
                elif code == DEFERRED_DEDENT_CODE:
                    levels = int(item[2:])
                    deferredDedents += levels
                elif code == EMPTY_CODE:
                    assert item == EMPTY_CODE
                else:
                    raise AssertionError("ERROR: Unknown control code <%s>" % code[1:])
            else:
                assert '\n' not in item, "rules must not contains newlines"
                if newlines:
                    yield '\n' + (newlines-1)*(''.join(indents).rstrip() + '\n')
                column += len(item)
                line += item
                newlines = 0
                prev_newline_was_soft = False
        assert len(indents) == deferredDedents, 'dedent missing'
        if newlines:
            yield '\n' + (newlines-1)*(''.join(indents).rstrip() + '\n')
        yield line

    # _eligibleToBeRule() {{{2
    @staticmethod
    def _eligibleToBeRule(rule):
        # Eligible to be a rule
        # Returns true if argument is a valid rule for Production.
        # Is a static method because rule may not be instance of Production.
        return (
            rule is None or
            is_str(rule) or
            isinstance(rule, Production) or
            hasattr(rule, 'docgen_rule') or
            hasattr(rule, 'docgen_templet') or
            hasattr(rule, 'docgen_template') or
            (   is_iterable(rule) and
                all([Production._eligibleToBeRule(r) for r in rule])
            )
        )

    # render() {{{2
    def render(
        self,
        indent=DEFAULT_INDENT,
        leaders=None,
        terminate=False
    ):
        """Convert production to a string

        :param indent:
            Indent productions will add this string to beginning of each line.
            If not specified, the indent is four spaces.

        :param leader:
            A dictionary of leaders. Leader productions will choose from these 
            leaders using the corresponding keys. If not specified, the default 
            leaders are used, which include '*' (maps to ' * ' and is used to 
            build C-style multiline comments), '#' (maps to '# ' and is used to 
            build Python style comments), and '//' (maps to '// ' and is used 
            for C++ style comments).
        :param terminate:
            A boolean that defaults to False. If true, a newline will be added 
            to the end of the output (normally that last newline is removed).
        :rtype: A string.
        """
        if leaders is None:
            leaders = DEFAULT_LEADERS
        return (
            strip_trailing_whitespace(''.join(self._filter(indent, leaders))) +
            ('\n' if terminate else '')
        )

    # __repr__() {{{2
    try:
        def __repr__(self):
            from kskutils import toStr
            return toStr(self)
    except ImportError:
        pass

    # __str__() {{{2
    def __str__(self):
        """Convert production to a string

        Use default indent and leaders.
        """
        return self.render()

    # expand() {{{2
    # make .expand() an alias for .render() for backward compatility
    expand = render

    # __bool__() {{{2
    def __bool__(self):
        return self.rule is not None

    # define __nonzero__ for backward compatibility with python2
    __nonzero__ = __bool__

# render {{{1
def render(arg):
    """Render production

    This intended only for debugging. It converts a production or a production 
    compatible object to a string.
    """
    return Production(arg).render()



import pytest
from inform import os_error
from textwrap import dedent
import subprocess
import os

# set various environment variables so avendesora uses local gpg key and config
# directory rather than the users.

os.environ['HOME'] = 'home'

# test_add() {{{1
def test_add():
    try:
        result = subprocess.check_output('avendesora help add'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Add a new account.

        Usage:
            avendesora [options] add [<template>]
            avendesora [options] a   [<template>]

        Options:
            -f <file>, --file <file>
                                    Add account to specified file.

        The default template is bank. The available templates are:
        bank, shell and website
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_browse() {{{1
def test_browse():
    try:
        result = subprocess.check_output('avendesora help browse'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Open account URL in web browser.

        Usage:
            avendesora [options] browse <account> [<key>]
            avendesora [options] b      <account> [<key>]

        Options:
            -b <browser>, --browser <browser>
                                    Open account in specified browser.

        The default browser is x. The available browsers are:
            f  firefox
            g  google-chrome
            t  torbrowser
            x  xdg-open
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_conceal() {{{1
def test_conceal():
    try:
        result = subprocess.check_output('avendesora help conceal'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Conceal text by encoding it.

        Usage:
            avendesora [options] conceal [<text>]
            avendesora [options] c       [<text>]

        Options:
            -e <encoding>, --encoding <encoding>
                                    Encoding used when concealing information.

        Possible encodings include:
            base64

        base64 (default):
            This encoding obscures but does not encrypt the text. It can
            protect text from observers that get a quick glance of the
            encoded text, but if they are able to capture it they can easily
            decode it.

        Though available as an option for convenience, you should not pass
        the text to be hidden as an argument as it is possible for others to
        examine the commands you run and their argument list. For any
        sensitive secret, you should simply run 'avendesora conceal' and
        then enter the secret text when prompted.
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_edit() {{{1
def test_edit():
    try:
        result = subprocess.check_output('avendesora help edit'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Edit an account.

        Usage:
            avendesora edit <account>
            avendesora e    <account>
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_find() {{{1
def test_find():
    try:
        result = subprocess.check_output('avendesora help find'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Find an account.

        Find accounts whose name contains the search text.

        Usage:
            avendesora find <text>
            avendesora f    <text>
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_help() {{{1
def test_help():
    try:
        result = subprocess.check_output('avendesora help help'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Give information about commands or other topics.

        Usage:
            avendesora help [<topic>]
            avendesora h    [<topic>]
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_initialize() {{{1
def test_initialize():
    try:
        result = subprocess.check_output('avendesora help initialize'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Create initial set of Avendesora files.

        Usage:
            avendesora initialize [--gpg-id <id>]... [options]
            avendesora I          [--gpg-id <id>]... [options]

        Options:
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.
            -h <path>, --gpg-home <path>
                                    GPG home directory (default is ~/.gnupg).

        Initial configuration and accounts files are created only if they
        do not already exist.  Existing files are not modified.
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_new() {{{1
def test_new():
    try:
        result = subprocess.check_output('avendesora help new'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Create new accounts file.

        Usage:
            avendesora new [--gpg-id <id>]... <name>
            avendesora N   [--gpg-id <id>]... <name>

        Options:
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.

        Creates a new accounts file. Accounts that share the same file share
        the same master password by default and, if the file is encrypted,
        can be decrypted by the same recipients.

        Generally you would create a new accounts file for each person or
        group with which you wish to share accounts. You would use separate
        files for passwords with different security domains. For example, a
        high-value passwords might be placed in an encrypted file that would
        only be placed highly on protected computers. Conversely, low-value
        passwords might be contained in perhaps an unencrypted file that is
        found on many computers.

        Add a '.gpg' extension to <name> to encrypt the file.
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_reveal() {{{1
def test_reveal():
    try:
        result = subprocess.check_output('avendesora help reveal'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Reveal concealed text.

        Transform concealed text to reveal its original form.

        Usage:
            avendesora [options] reveal [<text>]
            avendesora [options] r      [<text>]

        Options:
            -e <encoding>, --encoding <encoding>
                                    Encoding used when revealing information.

        Possible encodings include:
            base64

        base64 (default):
            This encoding obscures but does not encrypt the text. It can
            protect text from observers that get a quick glance of the
            encoded text, but if they are able to capture it they can easily
            decode it.

        Though available as an option for convenience, you should not pass
        the text to be revealed as an argument as it is possible for others
        to examine the commands you run and their argument list. For any
        sensitive secret, you should simply run 'avendesora reveal' and then
        enter the encoded text when prompted.
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_search() {{{1
def test_search():
    try:
        result = subprocess.check_output('avendesora help search'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Search accounts.

        Search for accounts whose values contain the search text.

        Usage:
            avendesora search <text>
            avendesora s      <text>
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_value() {{{1
def test_value():
    try:
        result = subprocess.check_output('avendesora help value'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Show an account value.

        Produce an account value. If the value is secret, it is produced only
        temporarily unless --stdout is specified.

        Usage:
            avendesora value [options] [--stdout | --clipboard] [<account> [<field>]]
            avendesora val   [options] [--stdout | --clipboard] [<account> [<field>]]
            avendesora v     [options] [--stdout | --clipboard] [<account> [<field>]]

        Options:
            -c, --clipboard         Write output to clipboard rather than stdout.
            -s, --stdout            Write output to the standard output without
                                    any annotation or protections.
            -v, --verbose           Add additional information to log file to
                                    help identify issues in account discovery.
            -t <title>, --title <title>
                                    Use account discovery on this title.
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_values() {{{1
def test_values():
    try:
        result = subprocess.check_output('avendesora help values'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Display all account values.

        Show all account values.

        Usage:
            avendesora values <account>
            avendesora vals   <account>
            avendesora V      <account>
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_version() {{{1
def test_version():
    try:
        result = subprocess.check_output('avendesora help version'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Display Avendesora version.

        Usage:
            avendesora version
    """).strip()
    assert result == bytes(expected, encoding='ascii')

# test_abraxas() {{{1
def test_abraxas():
    try:
        result = subprocess.check_output('avendesora help abraxas'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Avendesora generalizes and replaces Abraxas, its predecessor.  To
        transition from Abraxas to Avendesora, you will first need to
        upgrade Abraxas to version 1.8 or higher (use 'abraxas -v' to
        determine version). Then run:

            abraxas --export

        It will create a collection of Avendesora accounts files in
        ~/.config/abraxas/avendesora. You need to manually add these files
        to your list of accounts files in Avendesora. Say one such file in
        created: ~/.config/abraxas/avendesora/accounts.gpg.  This could be
        added to Avendesora as follows:

        1. create a symbolic link from
           ~/.config/avendesora/abraxas_accounts.gpg to
           ~/.config/abraxas/avendesora/accounts.gpg:

            cd ~/.config/avendesora
            ln -s ../abraxas/avendesora/accounts.gpg abraxas_accounts.gpg

        2. add abraxas_accounts.gpg to account_files list in .accounts_files.

        Now all of the Abraxas accounts contained in abraxas_accounts.gpg
        should be available though Avendesora and the various features of
        the account should operate as expected. However, secrets in accounts
        exported by Abraxas are no longer generated secrets. Instead, the
        actual secrets are placed in a hidden form in the exported accounts
        files.

        If you would like to enhance the imported accounts to take advantage
        of the new features of Avendesora, it is recommended that you do not
        manually modify the imported files. Instead, copy the account
        information to one of your own account files before modifying it.
        To avoid conflict, you must then delete the account from the
        imported file. To do so, create ~/.config/abraxas/do-not-export if
        it does not exist, then add the account name to this file, and
        reexport your accounts from Abraxas.
    """).strip()
    assert result == bytes(expected, encoding='utf8')

# test_discovery() {{{1
def test_discovery():
    try:
        result = subprocess.check_output('avendesora help discovery'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        If you do not give an account to 'avendesora value', Avendesora will
        try to determine the account by simply asking each account. An
        account can look at the window title, the user name, the host name,
        the working directory, and the environment variables to determine if
        it is suitable. If so, it nominates itself. If there is only one
        account nominated, that account is used. If there are multiple
        nominees, then a small window pops up allowing you to choose
        which account you wish to use.

        To configure an account to trigger when a particular window title is
        seen, use::

            discovery = RecognizeTitle(
                'Chase Online *',
                script='{username}{tab}{passcode}{return}'
            )

        The title is a glob string, meaning that '*' matches any combination
        of characters. The script describes what Avendesora should output
        when their is a match. In this case it outputs the username field,
        then a tab, then the passcode field, then a return.

        Matching window titles can be fragile, especially for websites
        because the titles can vary quite a bit across the site and over
        time. To accommodate this variation, you can give multiple glob
        patterns::

            discovery = RecognizeTitle(
                'CHASE Bank*',
                'Chase Online*',
                script='{username}{tab}{passcode}{return}'
            )

        If you use Firefox, you can install the 'Add URL to Window Title'
        extension. Doing so adds the website URL to the Firefox window
        title, which can make account discovery more robust. In this case
        you can use::

            discovery = RecognizeURL(
                'https://chaseonline.chase.com',
                'https://www.chase.com',
                script='{username}{tab}{passcode}{return}'
            )

        When giving the URL, anything specified must match and globbing is
        not supported. If you do not give the protocol, https is assumed.

        The following recognizers are available::

            RecognizeAll(<recognizer>..., [script=<script>])
            RecognizeAny(<recognizer>..., [script=<script>])
            RecognizeTitle(<title>..., [script=<script>])
            RecognizeURL(<title>..., [script=<script>, [name=<name>]])
            RecognizeHost(<host>..., [script=<script>])
            RecognizeUser(<user>..., [script=<script>])
            RecognizeCWD(<cwd>..., [script=<script>])
            RecognizeEnvVar(<name>, <value>, [script=<script>])

        RecognizeAll() and RecognizeAny() can be used to combine several
        recognizers. For example::

            discovery = RecognizeAll(
                RecognizeTitle('sudo *'),
                RecognizeUser('hhyde'),
                script='{passcode}{return}'
            )

        When account discovery fails it can be difficult to determine what
        is going wrong. When this occurs, you should first examine the log
        file. It should show you the window title and the recognized title
        components. You should first assure the title is as expected. If Add
        URL to Window Title generated the title, then the various title
        components should also be shown.  Then run Avendesora as follows::

            avendesora value --verbose --title '<title>'

        The title should be copied from the log file. The verbose option
        causes the result of each test to be included in the log file, so
        you can determine which recognizer is failing to trigger.

        If the recognizers are given in an array, all are tried. For
        example::

            discovery = [
                RecognizeURL(
                    'http://www.querty-forum.org',
                    script='admin{tab}{passcode}{return}',
                    name='admin',
                ),
                RecognizeURL(
                    'http://www.querty-forum.org',
                    script='thecaretaker{tab}{passcode}{return}',
                    name='thecaretaker',
                ),
            ]

        In this case, both recognizers recognize the same URL, thus they
        will both be offered for this site.  But each has a different
        script. The name allows the user to distinguish the available
        choices.
    """).strip()
    assert result == bytes(expected, encoding='utf8')

# test_entropy() {{{1
def test_entropy():
    try:
        result = subprocess.check_output('avendesora help entropy'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        A 4 word Avendesora password provides 53 bits of entropy, which
        seems like a lot, but NIST is recommending 80 bits for your most
        secure passwords.  So, how much is actually required. It is worth
        exploring this question.  Entropy is a measure of how hard the
        password is to guess. Specifically, it is the base two logarithm of
        the likelihood of guessing the password in a single guess. Every
        increase by one in the entropy represents a doubling in the
        difficulty of guessing your password. The actual entropy is hard to
        pin down, so generally we talk about the minimum entropy, which is
        the likelihood of an adversary guessing the password if he or she
        knows everything about the scheme used to generate the password but
        does not know the password itself.  So in this case the minimum
        entropy is the likelihood of guessing the password if it is known
        that we are using 4 space separated words as our pass phrase.  This
        is very easy to compute.  There are roughly 10,000 words in our
        dictionary, so if there was only one word in our pass phrase, the
        chance of guessing it would be one in 10,000 or 13 bits of entropy.
        If we used a two word pass phrase the chance of guessing it in a
        single guess is one in 10,000*10,000 or one in 100,000,000 or 26
        bits of entropy.

        The probability of guessing our pass phrase in one guess is not our
        primary concern. Really what we need to worry about is given a
        determined attack, how long would it take to guess the password. To
        calculate that, we need to know how fast our adversary could try
        guesses. If they are trying guesses by typing them in by hand, their
        rate is so low, say one every 10 seconds, that even a one word pass
        phrase may be enough to deter them.  Alternatively, they may have a
        script that automatically tries pass phrases through a login
        interface.  Again, generally the rate is relatively slow.  Perhaps
        at most the can get is 1000 tries per second. In this case they
        would be able to guess a one word pass phrase in 10 seconds and a
        two word pass phrase in a day, but a 4 word pass phrase would
        require 300,000 years to guess in this way.

        The next important thing to think about is how your password is
        stored by the machine or service you are logging into. The worst
        case situation is if they save the passwords in plain text. In this
        case if someone were able to break in to the machine or service,
        they could steal the passwords. Saving passwords in plain text is an
        extremely poor practice that was surprisingly common, but is
        becoming less common as companies start to realize their liability
        when their password files get stolen.  Instead, they are moving to
        saving passwords as hashes.  A hash is a transformation that is very
        difficult to reverse, meaning that if you have the password it is
        easy to compute its hash, but given the hash it is extremely
        difficult to compute the original password. Thus, they save the
        hashes (the transformed passwords) rather than the passwords. When
        you log in and provide your password, it is transformed with the
        hash and the result is compared against the saved hash. If they are
        the same, you are allowed in. In that way, your password is no
        longer available to thieves that break in.  However, they can still
        steal the file of hashed passwords, which is not as good as getting
        the plain text passwords, but it is still valuable because it allows
        thieves to greatly increase the rate that they can try passwords. If
        a poor hash was used to hash the passwords, then passwords can be
        tried at a very high rate.  For example, it was recently reported
        that password crackers were able to try 8 billion passwords per
        second when passwords were hashed with the MD5 algorithm. This would
        allow a 4 word pass phrase to be broken in 14 days, whereas a 6 word
        password would still require 4,000,000 years to break.  The rate for
        the more computational intensive sha512 hash was only 2,000
        passwords per second. In this case, a 4 word pass phrase would
        require 160,000 years to break.

        In most cases you have no control over how your passwords are stored
        on the machines or services that you log into.  Your best defense
        against the notoriously poor security practices of most sites is to
        always use a unique password for sites where you are not in control
        of the secrets.  For example, you might consider using the same pass
        phrase for you login password and the pass phrase for an ssh key on
        a machine that you administer, but never use the same password for
        two different websites unless you do not care if the content of
        those sites become public.

        So, if we return to the question of how much entropy is enough, you
        can say that for important passwords where you are in control of the
        password database and it is extremely unlikely to get stolen, then
        four randomly chosen words from a reasonably large dictionary is
        plenty.  If what the pass phrase is trying to protect is very
        valuable and you do not control the password database (ex., your
        brokerage account) you might want to follow the NIST recommendation
        and use 6 words to get 80 bits of entropy. If you are typing
        passwords on your work machine, many of which employ keyloggers to
        record your every keystroke, then no amount of entropy will protect
        you from anyone that has or gains access to the output of the
        keylogger.  In this case, you should consider things like one-time
        passwords or two-factor authentication. Or better yet, only access
        sensitive accounts from your home machine and not from any machine
        that you do not control.
    """).strip()
    assert result == bytes(expected, encoding='utf8')

# test_overview() {{{1
def test_overview():
    try:
        result = subprocess.check_output('avendesora help overview'.split())
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Avendesora is password utility that can store your account
        information and generate account passwords and produce them from the
        command line.  It can also be configured to autotype your username
        and password into the current window so that you can log in with a
        simple keystroke.

        Avendesora is capable of generating passwords (character-based
        passcodes) or pass phrases (word-based passcodes).  Pass phrases are
        generally preferred if you have a choice, but many websites will not
        take them.  The benefit of pass phrases is that they are relatively
        easy to remember and type, and they are very secure.  The pass
        phrases generated by Avendesora generally consist of four words,
        each word is drawn from a dictionary of 10,000 words.  Thus, even if
        a bad guy knew that four lower case words were being used for your
        pass phrase, there are still 10,000,000,000,000,000 possible
        combinations for him to try (this represents a minimum entropy of
        53 bits).  Using six words results in 80 bits of entropy, which
        meets the threshold recommended by NIST for the most secure pass
        phrases. For more on this, see 'avendesora help entropy' below.

        For another perspective on the attractiveness of pass phrases, see
        http://xkcd.com/936/.

        Unlike password vaults, Avendesora produces a highly unpredictable
        password from a master password and the name of the account for
        which the password is to be used. The process is completely
        repeatable. If you give the same master password and account name,
        you will get the same password. As such, the passwords do not have
        to be saved; instead they are regenerated on the fly.

        As a password generator, Avendesora provides three important
        advantages over conventional password vaults.  First, it allows
        groups of people to share access to accounts without having to
        securely share each password.  Instead, one member of the group
        creates a master password that is securely shared with the group
        once.  From then on any member of the group can create a new
        account, share the name of the account, and all members will know
        the password needed to access the account. The second advantage is
        that it opens up the possibility of using high-quality passwords for
        stealth accounts, which are accounts where you remember the name of
        the account but do not store any information about even the
        existence of the account on your computer.  With Avendesora, you
        only need to remember the name of the account and it will regenerate
        the password for you. This is perfect for your TrueCrypt hidden
        volume password.  Finally, by securely storing a small amount of
        infor‐ mation, perhaps on a piece of paper in your safe-deposit box,
        you can often recover most if not all of your passwords even if you
        somehow lose your accounts file. You can even recover passwords that
        were created after you created your backup. This is because
        Avendesora combines the master password with some easily recon‐
        structed information, such as the account name, to create the
        password. If you save the master password, the rest should be
        recoverable.

        To use it, one creates a file that contains information about each
        of his or her non-stealth accounts.  Among that information would be
        information that controls how the passwords are generated. This file
        is generally not encrypted, though you can encrypt it if you like).
        Another file is created that contains one or more master passwords.
        This file is always GPG encrypted.

        The intent is for these files to not include the passwords for your
        accounts.  Rather, the passwords are regenerated when needed from
        the account information and from the master password. This makes it
        easy to share passwords with others without having to pass the
        passwords back and forth.  It is only necessary to create a shared
        master password in advance. Then new passwords can be created on the
        fly by either party.
    """).strip()
    assert result == bytes(expected, encoding='utf8')




# Help
# Output a help topic.
#
# To Do:  Output the obscured text within a class instantiation. So rather than
# just outputting obscured text, output 'Hidden('obscured-text'). Also, accept
# obscured text within its class and try to decode it using the class name, so
# if GPG("ciphertext") then it will use GPG to decode it even if user does not
# provide the encoding.
#
# Also implement script class that uses a personal password. Call this Encrypt
# or PersonalEncrypt.

# License {{{1
# Copyright (C) 2016 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .command import Command
from .config import get_setting
from .obscure import Obscure
from .utilities import pager, two_columns
from inform import error, output
from textwrap import dedent

# HelpMessage base class {{{1
class HelpMessage:
    # get_name() {{{2
    @classmethod
    def get_name(cls):
        try:
            return cls.name.lower()
        except AttributeError:
            # consider converting lower to upper case transitions in __name__ to
            # dashes.
            return cls.__name__.lower()

    # topics {{{2
    @classmethod
    def topics(cls):
        for sub in cls.__subclasses__():
            yield sub

    # show {{{2
    @classmethod
    def show(cls, name=None):
        if name:
            command = Command.find(name)
            if command:
                return pager(command.help())
            for topic in cls.topics():
               if name == topic.get_name():
                   return pager(topic.help())
            error('topic not found.', culprit=name)
        else:
            cls.help()

    # summarize {{{2
    @classmethod
    def summarize(cls, width=16):
        summaries = []
        for topic in sorted(cls.topics(), key=lambda topic: topic.get_name()):
            summaries.append(two_columns(topic.get_name(), topic.DESCRIPTION))
        return '\n'.join(summaries)

    # help {{{2
    @classmethod
    def help(cls):
        output('Available commands:')
        output(Command.summarize())

        output('\nAvailable topics:')
        output(cls.summarize())


# Abraxas class {{{1
class Abraxas(HelpMessage):
    DESCRIPTION = "Exporting accounts from Abraxas"

    @staticmethod
    def help():
        text = dedent("""
            Avendesora generalizes and replaces Abraxas, its predecessor.  To
            transition from Abraxas to Avendesora, you will first need to
            upgrade Abraxas to version 1.8 or higher (use 'abraxas -v' to
            determine version). Then run:

                abraxas --export

            It will create a collection of Avendesora accounts files in
            ~/.config/abraxas/avendesora. You need to manually add these files
            to your list of accounts files in Avendesora. Say one such file in
            created: ~/.config/abraxas/avendesora/accounts.gpg.  This could be
            added to Avendesora as follows:

            1. create a symbolic link from
               ~/.config/avendesora/abraxas_accounts.gpg to
               ~/.config/abraxas/avendesora/accounts.gpg:

                cd ~/.config/avendesora
                ln -s ../abraxas/avendesora/accounts.gpg abraxas_accounts.gpg

            2. add abraxas_accounts.gpg to account_files list in .accounts_files.

            Now all of the Abraxas accounts contained in abraxas_accounts.gpg
            should be available though Avendesora and the various features of
            the account should operate as expected. However, secrets in accounts
            exported by Abraxas are no longer generated secrets. Instead, the
            actual secrets are placed in a hidden form in the exported accounts
            files.

            If you would like to enhance the imported accounts to take advantage
            of the new features of Avendesora, it is recommended that you do not
            manually modify the imported files. Instead, copy the account
            information to one of your own account files before modifying it.
            To avoid conflict, you must then delete the account from the
            imported file. To do so, create ~/.config/abraxas/do-not-export if
            it does not exist, then add the account name to this file, and
            reexport your accounts from Abraxas.
        """).strip()
        return text


# Discovery class {{{1
class Discovery(HelpMessage):
    DESCRIPTION = "Account discovery"

    @staticmethod
    def help():
        text = dedent("""
            If you do not give an account to 'avendesora value', Avendesora will
            try to determine the account by simply asking each account. An
            account can look at the window title, the user name, the host name,
            the working directory, and the environment variables to determine if
            it is suitable. If so, it nominates itself. If there is only one
            account nominated, that account is used. If there are multiple
            nominees, then a small window pops up allowing you to choose
            which account you wish to use.

            To configure an account to trigger when a particular window title is
            seen, use::

                discovery = RecognizeTitle(
                    'Chase Online *',
                    script='{username}{tab}{passcode}{return}'
                )

            The title is a glob string, meaning that '*' matches any combination
            of characters. The script describes what Avendesora should output
            when their is a match. In this case it outputs the username field,
            then a tab, then the passcode field, then a return.

            Matching window titles can be fragile, especially for websites
            because the titles can vary quite a bit across the site and over
            time. To accommodate this variation, you can give multiple glob
            patterns::

                discovery = RecognizeTitle(
                    'CHASE Bank*',
                    'Chase Online*',
                    script='{username}{tab}{passcode}{return}'
                )

            If you use Firefox, you can install the 'Add URL to Window Title'
            extension. Doing so adds the website URL to the Firefox window
            title, which can make account discovery more robust. In this case
            you can use::

                discovery = RecognizeURL(
                    'https://chaseonline.chase.com',
                    'https://www.chase.com',
                    script='{username}{tab}{passcode}{return}'
                )

            When giving the URL, anything specified must match and globbing is
            not supported. If you do not give the protocol, https is assumed.

            The following recognizers are available::

                RecognizeAll(<recognizer>..., [script=<script>])
                RecognizeAny(<recognizer>..., [script=<script>])
                RecognizeTitle(<title>..., [script=<script>])
                RecognizeURL(<title>..., [script=<script>, [name=<name>]])
                RecognizeHost(<host>..., [script=<script>])
                RecognizeUser(<user>..., [script=<script>])
                RecognizeCWD(<cwd>..., [script=<script>])
                RecognizeEnvVar(<name>, <value>, [script=<script>])

            RecognizeAll() and RecognizeAny() can be used to combine several
            recognizers. For example::

                discovery = RecognizeAll(
                    RecognizeTitle('sudo *'),
                    RecognizeUser('hhyde'),
                    script='{passcode}{return}'
                )

            When account discovery fails it can be difficult to determine what
            is going wrong. When this occurs, you should first examine the log
            file. It should show you the window title and the recognized title
            components. You should first assure the title is as expected. If Add
            URL to Window Title generated the title, then the various title
            components should also be shown.  Then run Avendesora as follows::

                avendesora value --verbose --title '<title>'

            The title should be copied from the log file. The verbose option
            causes the result of each test to be included in the log file, so
            you can determine which recognizer is failing to trigger.

            If the recognizers are given in an array, all are tried. For
            example::

                discovery = [
                    RecognizeURL(
                        'http://www.querty-forum.org',
                        script='admin{tab}{passcode}{return}',
                        name='admin',
                    ),
                    RecognizeURL(
                        'http://www.querty-forum.org',
                        script='thecaretaker{tab}{passcode}{return}',
                        name='thecaretaker',
                    ),
                ]

            In this case, both recognizers recognize the same URL, thus they
            will both be offered for this site.  But each has a different
            script. The name allows the user to distinguish the available
            choices.
        """).strip()
        return text


# Entropy class {{{1
class Entropy(HelpMessage):
    DESCRIPTION = "How much entropy is enough?"

    @staticmethod
    def help():
        text = dedent("""
            A 4 word Avendesora password provides 53 bits of entropy, which
            seems like a lot, but NIST is recommending 80 bits for your most
            secure passwords.  So, how much is actually required. It is worth
            exploring this question.  Entropy is a measure of how hard the
            password is to guess. Specifically, it is the base two logarithm of
            the likelihood of guessing the password in a single guess. Every
            increase by one in the entropy represents a doubling in the
            difficulty of guessing your password. The actual entropy is hard to
            pin down, so generally we talk about the minimum entropy, which is
            the likelihood of an adversary guessing the password if he or she
            knows everything about the scheme used to generate the password but
            does not know the password itself.  So in this case the minimum
            entropy is the likelihood of guessing the password if it is known
            that we are using 4 space separated words as our pass phrase.  This
            is very easy to compute.  There are roughly 10,000 words in our
            dictionary, so if there was only one word in our pass phrase, the
            chance of guessing it would be one in 10,000 or 13 bits of entropy.
            If we used a two word pass phrase the chance of guessing it in a
            single guess is one in 10,000*10,000 or one in 100,000,000 or 26
            bits of entropy.

            The probability of guessing our pass phrase in one guess is not our
            primary concern. Really what we need to worry about is given a
            determined attack, how long would it take to guess the password. To
            calculate that, we need to know how fast our adversary could try
            guesses. If they are trying guesses by typing them in by hand, their
            rate is so low, say one every 10 seconds, that even a one word pass
            phrase may be enough to deter them.  Alternatively, they may have a
            script that automatically tries pass phrases through a login
            interface.  Again, generally the rate is relatively slow.  Perhaps
            at most the can get is 1000 tries per second. In this case they
            would be able to guess a one word pass phrase in 10 seconds and a
            two word pass phrase in a day, but a 4 word pass phrase would
            require 300,000 years to guess in this way.

            The next important thing to think about is how your password is
            stored by the machine or service you are logging into. The worst
            case situation is if they save the passwords in plain text. In this
            case if someone were able to break in to the machine or service,
            they could steal the passwords. Saving passwords in plain text is an
            extremely poor practice that was surprisingly common, but is
            becoming less common as companies start to realize their liability
            when their password files get stolen.  Instead, they are moving to
            saving passwords as hashes.  A hash is a transformation that is very
            difficult to reverse, meaning that if you have the password it is
            easy to compute its hash, but given the hash it is extremely
            difficult to compute the original password. Thus, they save the
            hashes (the transformed passwords) rather than the passwords. When
            you log in and provide your password, it is transformed with the
            hash and the result is compared against the saved hash. If they are
            the same, you are allowed in. In that way, your password is no
            longer available to thieves that break in.  However, they can still
            steal the file of hashed passwords, which is not as good as getting
            the plain text passwords, but it is still valuable because it allows
            thieves to greatly increase the rate that they can try passwords. If
            a poor hash was used to hash the passwords, then passwords can be
            tried at a very high rate.  For example, it was recently reported
            that password crackers were able to try 8 billion passwords per
            second when passwords were hashed with the MD5 algorithm. This would
            allow a 4 word pass phrase to be broken in 14 days, whereas a 6 word
            password would still require 4,000,000 years to break.  The rate for
            the more computational intensive sha512 hash was only 2,000
            passwords per second. In this case, a 4 word pass phrase would
            require 160,000 years to break.

            In most cases you have no control over how your passwords are stored
            on the machines or services that you log into.  Your best defense
            against the notoriously poor security practices of most sites is to
            always use a unique password for sites where you are not in control
            of the secrets.  For example, you might consider using the same pass
            phrase for you login password and the pass phrase for an ssh key on
            a machine that you administer, but never use the same password for
            two different websites unless you do not care if the content of
            those sites become public.

            So, if we return to the question of how much entropy is enough, you
            can say that for important passwords where you are in control of the
            password database and it is extremely unlikely to get stolen, then
            four randomly chosen words from a reasonably large dictionary is
            plenty.  If what the pass phrase is trying to protect is very
            valuable and you do not control the password database (ex., your
            brokerage account) you might want to follow the NIST recommendation
            and use 6 words to get 80 bits of entropy. If you are typing
            passwords on your work machine, many of which employ keyloggers to
            record your every keystroke, then no amount of entropy will protect
            you from anyone that has or gains access to the output of the
            keylogger.  In this case, you should consider things like one-time
            passwords or two-factor authentication. Or better yet, only access
            sensitive accounts from your home machine and not from any machine
            that you do not control.
        """).strip()
        return text


# Overview class {{{1
class Overview(HelpMessage):
    DESCRIPTION = "overview of Avendesora"

    @staticmethod
    def help():
        text = dedent("""
            Avendesora is password utility that can store your account
            information and generate account passwords and produce them from the
            command line.  It can also be configured to autotype your username
            and password into the current window so that you can log in with a
            simple keystroke.

            Avendesora is capable of generating passwords (character-based
            passcodes) or pass phrases (word-based passcodes).  Pass phrases are
            generally preferred if you have a choice, but many websites will not
            take them.  The benefit of pass phrases is that they are relatively
            easy to remember and type, and they are very secure.  The pass
            phrases generated by Avendesora generally consist of four words,
            each word is drawn from a dictionary of 10,000 words.  Thus, even if
            a bad guy knew that four lower case words were being used for your
            pass phrase, there are still 10,000,000,000,000,000 possible
            combinations for him to try (this represents a minimum entropy of
            53 bits).  Using six words results in 80 bits of entropy, which
            meets the threshold recommended by NIST for the most secure pass
            phrases. For more on this, see 'avendesora help entropy' below.

            For another perspective on the attractiveness of pass phrases, see
            http://xkcd.com/936/.

            Unlike password vaults, Avendesora produces a highly unpredictable
            password from a master password and the name of the account for
            which the password is to be used. The process is completely
            repeatable. If you give the same master password and account name,
            you will get the same password. As such, the passwords do not have
            to be saved; instead they are regenerated on the fly.

            As a password generator, Avendesora provides three important
            advantages over conventional password vaults.  First, it allows
            groups of people to share access to accounts without having to
            securely share each password.  Instead, one member of the group
            creates a master password that is securely shared with the group
            once.  From then on any member of the group can create a new
            account, share the name of the account, and all members will know
            the password needed to access the account. The second advantage is
            that it opens up the possibility of using high-quality passwords for
            stealth accounts, which are accounts where you remember the name of
            the account but do not store any information about even the
            existence of the account on your computer.  With Avendesora, you
            only need to remember the name of the account and it will regenerate
            the password for you. This is perfect for your TrueCrypt hidden
            volume password.  Finally, by securely storing a small amount of
            infor‐ mation, perhaps on a piece of paper in your safe-deposit box,
            you can often recover most if not all of your passwords even if you
            somehow lose your accounts file. You can even recover passwords that
            were created after you created your backup. This is because
            Avendesora combines the master password with some easily recon‐
            structed information, such as the account name, to create the
            password. If you save the master password, the rest should be
            recoverable.

            To use it, one creates a file that contains information about each
            of his or her non-stealth accounts.  Among that information would be
            information that controls how the passwords are generated. This file
            is generally not encrypted, though you can encrypt it if you like).
            Another file is created that contains one or more master passwords.
            This file is always GPG encrypted.

            The intent is for these files to not include the passwords for your
            accounts.  Rather, the passwords are regenerated when needed from
            the account information and from the master password. This makes it
            easy to share passwords with others without having to pass the
            passwords back and forth.  It is only necessary to create a shared
            master password in advance. Then new passwords can be created on the
            fly by either party.
        """).strip()
        return text





# -*- coding: utf-8 -*-
import logging
import os
import warnings
import tempfile
import shutil
import json

from tarfile import TarFile
from pkgutil import get_data
from io import BytesIO
from contextlib import closing

from dateutil.tz import tzfile

__all__ = ["gettz", "gettz_db_metadata", "rebuild"]

ZONEFILENAME = "dateutil-zoneinfo.tar.gz"
METADATA_FN = 'METADATA'

# python2.6 compatability. Note that TarFile.__exit__ != TarFile.close, but
# it's close enough for python2.6
tar_open = TarFile.open
if not hasattr(TarFile, '__exit__'):
    def tar_open(*args, **kwargs):
        return closing(TarFile.open(*args, **kwargs))


class tzfile(tzfile):
    def __reduce__(self):
        return (gettz, (self._filename,))


def getzoneinfofile_stream():
    try:
        return BytesIO(get_data(__name__, ZONEFILENAME))
    except IOError as e:  # TODO  switch to FileNotFoundError?
        warnings.warn("I/O error({0}): {1}".format(e.errno, e.strerror))
        return None


class ZoneInfoFile(object):
    def __init__(self, zonefile_stream=None):
        if zonefile_stream is not None:
            with tar_open(fileobj=zonefile_stream, mode='r') as tf:
                # dict comprehension does not work on python2.6
                # TODO: get back to the nicer syntax when we ditch python2.6
                # self.zones = {zf.name: tzfile(tf.extractfile(zf),
                #               filename = zf.name)
                #              for zf in tf.getmembers() if zf.isfile()}
                self.zones = dict((zf.name, tzfile(tf.extractfile(zf),
                                                   filename=zf.name))
                                  for zf in tf.getmembers()
                                  if zf.isfile() and zf.name != METADATA_FN)
                # deal with links: They'll point to their parent object. Less
                # waste of memory
                # links = {zl.name: self.zones[zl.linkname]
                #        for zl in tf.getmembers() if zl.islnk() or zl.issym()}
                links = dict((zl.name, self.zones[zl.linkname])
                             for zl in tf.getmembers() if
                             zl.islnk() or zl.issym())
                self.zones.update(links)
                try:
                    metadata_json = tf.extractfile(tf.getmember(METADATA_FN))
                    metadata_str = metadata_json.read().decode('UTF-8')
                    self.metadata = json.loads(metadata_str)
                except KeyError:
                    # no metadata in tar file
                    self.metadata = None
        else:
            self.zones = dict()
            self.metadata = None


# The current API has gettz as a module function, although in fact it taps into
# a stateful class. So as a workaround for now, without changing the API, we
# will create a new "global" class instance the first time a user requests a
# timezone. Ugly, but adheres to the api.
#
# TODO: deprecate this.
_CLASS_ZONE_INSTANCE = list()


def gettz(name):
    if len(_CLASS_ZONE_INSTANCE) == 0:
        _CLASS_ZONE_INSTANCE.append(ZoneInfoFile(getzoneinfofile_stream()))
    return _CLASS_ZONE_INSTANCE[0].zones.get(name)


def gettz_db_metadata():
    """ Get the zonefile metadata

    See `zonefile_metadata`_

    :returns: A dictionary with the database metadata
    """
    if len(_CLASS_ZONE_INSTANCE) == 0:
        _CLASS_ZONE_INSTANCE.append(ZoneInfoFile(getzoneinfofile_stream()))
    return _CLASS_ZONE_INSTANCE[0].metadata





#
# Templet Class
#
# Accepts syntax descriptions that are reminiscent of usage strings and converts 
# them into docgen productions.
#

# Imports {{{1
from __future__ import print_function
from .production import Production
from .predefined import Alternates, Optional, Join, Cull, Tab, Newline
from .tabfiles import TabFiles
import ply.lex, ply.yacc
import collections
import os

# Globals {{{1
Tabstop = None
IRs = {}    # Cache for the intermediate representations
            # (avoids compiling the same templet over and over)

# Utilities {{{1
def toStr(arg):
    if type(arg) == list:
        return '[%s]' % ', '.join([toStr(a) for a in arg])
    else:
        return str(arg)

def flatten(iterable):
    for each in iterable:
        if isinstance(each, list):
            for e in flatten(each):
                yield e
        else:
            yield each


# Attributes class {{{1
# The user passes in attribute values to the templet using args and kwargs. The 
# Attributes class is responsible for searching through potential attribute 
# values and finding the one with the right name. The kwargs are searched 
# directly. Items in args may either be dictionaries or objects, and the first 
# item or attribute with the right name is used. This class also handles 
# hierarchical attributes; attributes of the form 'container.name' that may be 
# nested arbitrarily deeply.

class Attributes(object):
    def __init__(self, nullify, args, kwargs):
        self.nullify = nullify
        self.args = args
        self.kwargs = kwargs

    def getValue(self, key, indices, _collection = None):
        if '.' in key:
            topkey, subkey = key.split('.', 1)
            collection = self._get(topkey, _collection)
            if collection:
                if type(collection) == list:
                    val = self.getValue(
                        key=subkey,
                        indices=indices[1:],
                        _collection=indices[0].select(collection)
                    )
                else:
                    val = self.getValue(subkey, indices, collection)
            else:
                val = None
        else:
            if indices:
                val = self._get(key, _collection)
                for i in range(len(indices)):
                    val = indices[i].select(val)
            else:
                val = self._get(key, _collection)
        return val

    def _get(self, key, collection):
        # Get the value from a collection of values accommodating the fact that 
        # the collection may be a dictionary or an object, and accommodating 
        # both top level keys (search the collections given when this class was 
        # instantiated) or a lower level key (search the collection specified 
        # by the higher level keys).

        if collection is None:
            # this is a top-level attribute
            collections = [self.kwargs] + list(self.args)
        else:
            # this is a lower-level attribute of the given collection
            collections = [collection]
        for arg in collections:
            try:
                return getattr(arg, key)
            except (AttributeError, TypeError):
                pass
            try:
                return arg[key]
            except (KeyError, AttributeError, TypeError):
                pass
        if self.nullify:
            return None
        raise TempletError('%s: not found.' % key)

# Index class {{{1
class Index(object):
    # Class that handles accessing of array values for joins. One object of 
    # this class is intended to handle one dimension of perhaps 
    # a multi-dimensional access. These objects are always expected to be 
    # passed around in an array so as to allow support of multi-dimensional 
    # indexing.
    def __init__(self, initialValue=0):
        self.index = initialValue
        self.arrayFound = False

    def increment(self):
        if not self.arrayFound:
            raise IndexError
        self.index += 1

    def select(self, data):
        if type(data) is not str and isinstance(data, collections.Iterable):
            self.arrayFound = True
            return data[self.index]
        else:
            return data

    def __repr__(self):
        return 'Index(%s)' % (self.index)

# Deferred classes {{{1
# Deferred classes are used as the intermediate representation (IR) for the 
# parser, meaning that that input text is broken down into its recognizable 
# parts and converted into a hierarchical collection of strings, lists, and 
# objects of the Deferred class.  This is done so as to defer the conversion 
# into docgen productions until the join processing can occur. In join 
# processing, subelements are evaluated repeatedly while iterating over the 
# items of a list.

# Deferred class {{{2
class Deferred(object):
    def __init__(self, arg=None):
        self.arg = arg

    def expand(self, attrs, index=()):
        rule = []
        for arg in flatten([self.arg]):
            if isinstance(arg, Deferred):
                rule.append(arg.expand(attrs, index))
            else:
                assert isinstance(arg, (type(None), str, Production)) , ('found %s' % type(arg))
                rule.append(arg)
        return self._finalize(rule)

    def _finalize(self, rule):
        return Production(rule)

    def __str__(self):
        return '%s' % toStr(self.arg)


# Deferred Attribute class {{{2
class DeferredAttribute(Deferred):
    def expand(self, attrs, index=()):
        return attrs.getValue(self.arg, index)

    def _finalize(self, rules):
        return rules

    def __str__(self):
        return '<%s>' % self.arg


# Deferred Alternates class {{{2
class DeferredAlternates(Deferred):
    def _finalize(self, rules):
        return Alternates(rules)

    def __str__(self):
        return '{%s}' % '|'.join([toStr(arg) for arg in self.arg])


# Deferred Optional class {{{2
class DeferredOptional(Deferred):
    def _finalize(self, rules):
        assert len(rules) == 1
        rule = rules[0]
        return Optional(rule, alt="")

    def __str__(self):
        return '{[%s]}' % toStr(self.arg)


# Deferred Join class {{{2
class DeferredJoin(Deferred):
    def __init__(self, arg, sep):
        self.arg = arg
        self.sep = sep

    def expand(self, attrs, indices=()):
        arg = self.arg
        rules = []
        innerIndex = Index()
        indices = indices + (innerIndex,)
        try:
            while(True):
                if isinstance(arg, Deferred):
                    rules.append(arg.expand(attrs, indices))
                else:
                    assert type(arg) in [type(None), str], ('found %s' % type(arg))
                    rules.append(arg)
                innerIndex.increment()
        except IndexError:
            pass
        return self._finalize(rules)

    def _finalize(self, rules):
        return Join(rules, self.sep)

    def __str__(self):
        return '{%s}%s...' % (toStr(self.arg), self.sep)


# Deferred Tab class {{{2
class DeferredTab(Deferred):
    def _finalize(self, rules):
        return Tab(Tabstop)

    def __str__(self):
        return 'Tab(%s)' % Tabstop


# Exceptions {{{1
class TempletError(Exception):
    def __init__(self, message, lexer=None, lexpos=None):
        self.message = message
        if lexer:
            bol = lexer.lexdata.rfind('\n', 0, lexpos) + 1
            if bol < 0:
                bol = 0
            eol = lexer.lexdata.find('\n', lexpos)
            if eol < 0:
                eol = len(lexer.lexdata)
            self.line = lexer.lexdata[bol:eol]
            self.col = (lexpos - bol)
            self.row = lexer.lexdata[0:lexpos].count('\n') + 1

    def __str__(self):
        "Return a formatted error message."
        try:
            location = "%s\n%s^\n%s: " % (
                self.line, self.col*' ', self.row
            )
        except AttributeError:
            location = ""
        return location + self.message

# Lexer {{{1
reserved = {}
tokens = [
    'ID',
    'LBRACE',
    'RBRACE',
    'LANGLE',
    'RANGLE',
    'LBRACKET',
    'RBRACKET',
    'PIPE',
    'DOTS',
    'TAB',
    'NEWLINE',
    'CHAR',
]
literals = ''

# Regular expressions that define identifiers
def t_ID(t):
    r'\b\w+(\.\w+)*\b'
    t.type = reserved.get(t.value, 'ID')    # check for reserved words
    return t

# Regular expressions that define the symbols
# Could not put these in literals because I wanted to allow excaping
t_LBRACE = r"(?<!\\){"      # matches { if not preceded by \
t_RBRACE = r"(?<!\\)}"      # matches } if not preceded by \
t_LANGLE = r"(?<!\\)<"      # matches < if not preceded by \
t_RANGLE = r"(?<!\\)>"      # matches > if not preceded by \
t_LBRACKET = r"(?<!\\)\["   # matches [ if not preceded by \
t_RBRACKET = r"(?<!\\)\]"   # matches ] if not preceded by \
t_PIPE = r"(?<!\\)\|"       # matches | if not preceded by \
t_DOTS = r"\.\.\."          # matches ...
t_TAB = r"\t"               # matches tab
t_NEWLINE = r"\n"           # matches newline
t_CHAR = r"."               # matches any character not previously matched
                            # ply orders tokens from longest to shortest, so 
                            # this one must be the shortest

# Error handler (this should never happen)
def t_error(t):
    c = t.value[0]
    lexer, lexpos = t.lexer, t.lexpos
    t.lexer.skip(1)
    raise TempletError("illegal character '%s'." % c, lexer, lexpos)

# Parser {{{1
def p_template(p):
    "template : alternatives"
    if len(p[1]) > 1:
        p[0] = DeferredAlternates(p[1])
    else:
        p[0] = Deferred(p[1])

def p_simple_template(p):
    "template : text"
    p[0] = Deferred(p[1])

def p_alternatives(p):
    "alternatives : alternatives PIPE terms"
    p[0] = p[1] + [Deferred(p[3])]

def p_alternatives_last1(p):
    "alternatives : terms"
    p[0] = [Deferred(p[1])]

def p_alternatives_last2(p):
    "alternatives : text terms"
    p[0] = [Deferred(['%s' % p[1]] + p[2])]

def p_terms(p):
    "terms : terms term"
    p[0] = p[1] + p[2]

def p_terms_last(p):
    "terms : term"
    p[0] = p[1]

def p_term(p):
    """
    term : atom text
    """
    p[0] = [p[1], '%s' % p[2]]

def p_naked_term(p):
    """
    term : atom
    """
    p[0] = [p[1]]

def p_atom(p):
    """
    atom : attribute
         | group
         | option
         | join
         | tab
         | newline
    """
    p[0] = Deferred(p[1])

def p_attribute(p):
    "attribute : LANGLE ID RANGLE"
    p[0] = DeferredAttribute(p[2])

def p_group(p):
    "group : LBRACE template RBRACE"
    p[0] = p[2]

def p_option(p):
    "option : LBRACKET template RBRACKET"
    p[0] = DeferredOptional(p[2])

def p_join(p):
    "join : atom text DOTS"
    p[0] = DeferredJoin(p[1], p[2])

def p_tab(p):
    "tab : TAB"
    p[0] = DeferredTab()

def p_newline(p):
    "newline : NEWLINE"
    p[0] = Newline('!')

def p_text(p):
    """
    text : text CHAR
         | text ID
    """
    p[0] = p[1] + (p[2] if p[2] != '\\' else '')

def p_text_last(p):
    """
    text : CHAR
         | ID
    """
    p[0] = p[1] if p[1] != '\\' else ''

# Error rule for syntax errors
def p_error(p):
    if p:
        raise TempletError("syntax error at '%s'." % (p.value), p.lexer, p.lexpos)
    else:
        raise TempletError("premature end of content.")

# Parse {{{1
tabfiles = TabFiles(__file__)
Lexer = ply.lex.lex(
    debug=False
  , optimize=True
  , lextab=tabfiles.lextab
  , outputdir=tabfiles.directory
)
Parser = ply.yacc.yacc(
    debug=False
  , optimize=True
  , method="SLR"
  , tabmodule=tabfiles.parsetab
  , outputdir=tabfiles.directory
)
def parse(templet, attrs):
    if templet not in IRs:
        IRs[templet] = Parser.parse(templet, lexer=Lexer)
    return IRs[templet].expand(attrs)

# Templet class {{{1
class Templet(Production):
    def __init__(self, templet, *args, **kwargs):
        self.add_rule(self._elaborate(self, templet, *args, **kwargs))

    def add_templet(self, templet, *args, **kwargs):
        self.rule = self._clean(self._elaborate(self, templet, *args, **kwargs))

    @staticmethod
    def _elaborate(rule, templet, *args, **kwargs):
        global Tabstop
        Tabstop = kwargs.pop('tabstop', None)
        nullify = kwargs.pop('nullify', None)
        attrs = Attributes(nullify, args + (rule,), kwargs)
        if type(templet) == str:
            return parse(templet, attrs)
        elif isinstance(templet, collections.Iterable):
            for each in templet:
                result = parse(each, attrs)
                if result:
                    return result
        raise AssertionError("Invalid type for templet value")
